/*
 *  TOPPERS/ASP Kernel
 *      Toyohashi Open Platform for Embedded Real-Time Systems/
 *      Advanced Standard Profile Kernel
 * 
 *  Copyright (C) 2008-2011 by Embedded and Real-Time Systems Laboratory
 *              Graduate School of Information Science, Nagoya Univ., JAPAN
 *  Copyright (C) 2015-2016 by TOPPERS PROJECT Educational Working Group.
 * 
 *  嵭Ԥϡʲ(1)(4)ξ˸¤ꡤܥեȥ
 *  ܥեȥѤΤޤࡥʲƱˤѡʣ
 *  ѡۡʰʲѤȸƤ֡ˤ뤳Ȥ̵ǵ롥
 *  (1) ܥեȥ򥽡ɤηѤˤϡ嵭
 *      ɽѾ浪Ӳ̵ݾڵ꤬Τޤޤηǥ
 *      ˴ޤޤƤ뤳ȡ
 *  (2) ܥեȥ򡤥饤֥ʤɡ¾Υեȥȯ˻
 *      ѤǤǺۤˤϡۤȼɥȡ
 *      ԥޥ˥奢ʤɡˤˡ嵭ɽѾ浪Ӳ
 *      ̵ݾڵǺܤ뤳ȡ
 *  (3) ܥեȥ򡤵Ȥ߹ʤɡ¾Υեȥȯ˻
 *      ѤǤʤǺۤˤϡΤ줫ξ
 *      ȡ
 *    (a) ۤȼɥȡѼԥޥ˥奢ʤɡˤˡ嵭
 *        ɽѾ浪Ӳ̵ݾڵǺܤ뤳ȡ
 *    (b) ۤη֤̤ˡˤäơTOPPERSץȤ
 *        𤹤뤳ȡ
 *  (4) ܥեȥѤˤľŪޤϴŪ뤤ʤ»
 *      ⡤嵭ԤTOPPERSץȤդ뤳ȡ
 *      ޤܥեȥΥ桼ޤϥɥ桼Τʤ
 *      ͳ˴Ťᤫ⡤嵭ԤTOPPERSץȤ
 *      դ뤳ȡ
 * 
 *  ܥեȥϡ̵ݾڤ󶡤ƤΤǤ롥嵭Ԥ
 *  TOPPERSץȤϡܥեȥ˴ؤơλŪ
 *  ФŬޤơʤݾڤԤʤޤܥեȥ
 *  ѤˤľŪޤϴŪʤ»˴ؤƤ⡤
 *  Ǥʤ
 * 
 *  @(#) $Id: i2c.c 698 2016-02-10 12:05:16Z roi $
 */
/*
 * 
 *  I2Cɥ饤дؿ
 *
 */
#include "kernel_impl.h"
#include <t_syslog.h>
#include <t_stdlib.h>
#include <string.h>
#include <sil.h>
#include <target_syssvc.h>
#include "device.h"
#include "i2c.h"

#define sil_orw_mem(a, b)		sil_wrw_mem((a), sil_rew_mem(a) | (b))
#define sil_andw_mem(a, b)		sil_wrw_mem((a), sil_rew_mem(a) & ~(b))
#define sil_modw_mem(a, b, c)	sil_wrw_mem((a), (sil_rew_mem(a) & (~b)) | (c))

/*
 *  I2CݡID֥åФΥޥ
 */
#define INDEX_I2C(i2cid)        ((uint_t)((i2cid) - 1))

/*
 *  I2CAF
 */ 
#define GPIO_AF4_I2C1          ((uint8_t)0x04)  /* I2C1 Alternate Function mapping */
#define GPIO_AF4_I2C2          ((uint8_t)0x04)  /* I2C2 Alternate Function mapping */
#define GPIO_AF4_I2C3          ((uint8_t)0x04)  /* I2C3 Alternate Function mapping */
#ifdef TOPPERS_STM32F401_NUCLEO
#define GPIO_AFN_I2C2          ((uint8_t)0x09)  /* I2C2 Alternate Function mapping */
#define GPIO_AFN_I2C3          ((uint8_t)0x09)  /* I2C3 Alternate Function mapping */
#define GPIO_I2C3_SDA          TADR_GPIOB_BASE
#define I2C3_SDA               4
#else
#define GPIO_AFN_I2C2          ((uint8_t)0x04)  /* I2C2 Alternate Function mapping */
#define GPIO_AFN_I2C3          ((uint8_t)0x04)  /* I2C3 Alternate Function mapping */
#define GPIO_I2C3_SDA          TADR_GPIOC_BASE
#define I2C3_SDA               9
#endif
#ifdef TOPPERS_STM32F4_DISCOVERY
#define I2C2_SDA               11
#else
#define I2C2_SDA               3
#endif

#define I2C_TIMEOUT_FLAG            35			/* 35 ms */
#define I2C_TIMEOUT_BUSY_FLAG       10000		/* 10 s  */

#define I2C_FREQRANGE(_PCLK_)       ((_PCLK_)/1000000)

#define I2C_RISE_TIME(_FREQ_, _SPEED_)            (((_SPEED_) <= 100000) ? ((_FREQ_) + 1) : ((((_FREQ_) * 300) / 1000) + 1))
#define I2C_SPEED_STANDARD(_PCLK_, _SPEED_)       (((((_PCLK_)/((_SPEED_) << 1)) & I2C_CCR_CCR) < 4)? 4:((_PCLK_) / ((_SPEED_) << 1)))
#define I2C_SPEED_FAST(_PCLK_, _SPEED_, _DCYCLE_) (((_DCYCLE_) == I2C_DUTYCYCLE_2)? ((_PCLK_) / ((_SPEED_) * 3)) : (((_PCLK_) / ((_SPEED_) * 25)) | I2C_DUTYCYCLE_16_9))
#define I2C_SPEED(_PCLK_, _SPEED_, _DCYCLE_)      (((_SPEED_) <= 100000)? (I2C_SPEED_STANDARD((_PCLK_), (_SPEED_))) : \
                                                                  ((I2C_SPEED_FAST((_PCLK_), (_SPEED_), (_DCYCLE_)) & I2C_CCR_CCR) == 0)? 1 : \
                                                                  ((I2C_SPEED_FAST((_PCLK_), (_SPEED_), (_DCYCLE_))) | I2C_CCR_FS))

#define I2C_7BIT_ADD_WRITE(_ADDR_)  ((uint8_t)((_ADDR_) & (~I2C_OAR1_ADD0)))
#define I2C_7BIT_ADD_READ(_ADDR_)   ((uint8_t)((_ADDR_) | I2C_OAR1_ADD0))

#define I2C_10BIT_ADDRESS(_ADDR_)   ((uint8_t)(((_ADDR_) & 0x00FF)))
#define I2C_10BIT_HEADWRITE(_ADDR_) ((uint8_t)(((((_ADDR_) & 0x0300) >> 7) | 0xF0)))
#define I2C_10BIT_HEADREAD(_ADDR_)  ((uint8_t)(((((_ADDR_) & 0x0300) >> 7) | 0xF1)))

#define I2C_MEM_ADD_MSB(_ADDR_)     ((uint8_t)(((_ADDR_) & 0xFF00) >> 8))
#define I2C_MEM_ADD_LSB(_ADDR_)     ((uint8_t)((_ADDR_) & 0x00FF))


/*
 *  I2Cݡơ֥
 */
static const I2C_PortControlBlock i2c_pcb[NUM_I2CPORT] = {
  {	TADR_I2C1_BASE,
	(TADR_RCC_BASE+TOFF_RCC_AHB1ENR),  RCC_AHB1ENR_GPIOBEN, RCC_AHB1ENR_GPIOBEN,
	(TADR_RCC_BASE+TOFF_RCC_APB1ENR),  RCC_APB1ENR_I2C1EN,
	(TADR_RCC_BASE+TOFF_RCC_APB1RSTR), RCC_APB1RSTR_I2C1RST,
	TADR_GPIOB_BASE, TADR_GPIOB_BASE,
	8, GPIO_AF4_I2C1, 9, GPIO_AF4_I2C1 },

  {	TADR_I2C2_BASE,
	(TADR_RCC_BASE+TOFF_RCC_AHB1ENR),  RCC_AHB1ENR_GPIOBEN, RCC_AHB1ENR_GPIOBEN,
	(TADR_RCC_BASE+TOFF_RCC_APB1ENR),  RCC_APB1ENR_I2C2EN,
	(TADR_RCC_BASE+TOFF_RCC_APB1RSTR), RCC_APB1RSTR_I2C2RST,
	TADR_GPIOB_BASE, TADR_GPIOB_BASE,
	10, GPIO_AF4_I2C2, I2C2_SDA, GPIO_AFN_I2C2 },

  {	TADR_I2C3_BASE,
	(TADR_RCC_BASE+TOFF_RCC_AHB1ENR),  RCC_AHB1ENR_GPIOAEN, RCC_AHB1ENR_GPIOBEN,
	(TADR_RCC_BASE+TOFF_RCC_APB1ENR),  RCC_APB1ENR_I2C3EN,
	(TADR_RCC_BASE+TOFF_RCC_APB1RSTR), RCC_APB1RSTR_I2C3RST,
	TADR_GPIOA_BASE, GPIO_I2C3_SDA,
	8, GPIO_AF4_I2C3, I2C3_SDA, GPIO_AFN_I2C3 }
};


/*
 * I2C ϥɥ
 */
static I2C_Handle_t I2cHandle[NUM_I2CPORT];

/*
 *  SR1쥸åԤ
 *  parameter1 hi2c:    I2CϥɥؤΥݥ
 *  parameter2 Flag:    Ԥե饰
 *  parameter3 Timeout: Ԥ(ms)
 *  return              ｪλE_OK
 */
static ER
i2c_sr1flagsetwait(I2C_Handle_t *hi2c, uint32_t Flag, uint32_t Timeout)
{
	uint32_t tick = 0;
	uint32_t ctime = Timeout * 1000;

	while((sil_rew_mem((uint32_t *)(hi2c->base+TOFF_I2C_SR1)) & Flag) == 0){
		if(ctime == 0 || tick > ctime){
			hi2c->status= I2C_STATUS_READY;
			return E_TMOUT;
		}
        sil_dly_nse(1000);
		tick++;
	}
	return E_OK;
}

/*
 *  BUS READYԤ
 *  parameter1 hi2c:    I2CϥɥؤΥݥ
 *  parameter2 Timeout: Ԥ(ms)
 *  return              ｪλE_OK
 */
static ER
i2c_busreadywait(I2C_Handle_t *hi2c, uint32_t Timeout)
{
	uint32_t tick = 0;
	uint32_t ctime = Timeout * 1000;

    while((sil_rew_mem((uint32_t *)(hi2c->base+TOFF_I2C_SR2)) & I2C_SR2_BUSY) != 0){
		if(ctime == 0 || tick > ctime){
			hi2c->status= I2C_STATUS_READY;
			return E_TMOUT;
		}
        sil_dly_nse(1000);
		tick++;
	}
	return E_OK;
}

/*
 * ɥ쥹ϥե饰Ԥ
 *  parameter1 hi2c:    I2CϥɥؤΥݥ
 *  parameter2 Flag:    Ԥե饰
 *  parameter3 Timeout: Ԥ(ms)
 *  return              ｪλE_OK
 */
static ER
i2c_addrflagwait(I2C_Handle_t *hi2c, uint32_t Flag, uint32_t Timeout)
{
	uint32_t tick = 0;
	uint32_t ctime = Timeout * 1000;

	while((sil_rew_mem((uint32_t *)(hi2c->base+TOFF_I2C_SR1)) & Flag) == 0){
		/*
		 *  ΥåեȽ
		 */
		if((sil_rew_mem((uint32_t *)(hi2c->base+TOFF_I2C_SR1)) & I2C_SR1_AF) != 0){
			sil_orw_mem((uint32_t *)(hi2c->base+TOFF_I2C_CR1), I2C_CR1_STOP);
			sil_wrw_mem((uint32_t *)(hi2c->base+TOFF_I2C_SR1), ~I2C_SR1_AF);
			hi2c->ErrorCode = I2C_ERROR_AF;
			hi2c->status= I2C_STATUS_READY;
			return E_SYS;
		}
		if(ctime == 0 || tick > ctime){
			hi2c->status= I2C_STATUS_READY;

			return E_TMOUT;
		}
		sil_dly_nse(1000);
		tick++;
	}
	return E_OK;
}

/*
 *  ɥ쥹ե饰ꥢ
 *  parameter1 hi2c:   I2CϥɥؤΥݥ
 */
static void
i2c_clear_addr(I2C_Handle_t *hi2c)
{
	volatile uint32_t tmp;

	tmp = sil_rew_mem((uint32_t *)(hi2c->base+TOFF_I2C_SR1));
	tmp = sil_rew_mem((uint32_t *)(hi2c->base+TOFF_I2C_SR2));
	(void)(tmp);
}

/*
 *  I2CžԤ
 *  parameter1 hi2c:   I2CϥɥؤΥݥ
 *  parameter2 Timeout:ॢȻ(ms)
 */
static ER
i2c_transwait(I2C_Handle_t *hi2c, uint32_t Timeout)
{
	uint32_t tick = 0;

	while(hi2c->status != I2C_STATUS_READY){
		if(tick >= Timeout)
			return E_TMOUT;
		if(hi2c->Init.semid != 0)
			twai_sem(hi2c->Init.semid, 1);
		else
			dly_tsk(1);
		tick++;
	}
	return E_OK;
}

/*
 *  ޥ ǥХɥ쥹ȥꥢɥ쥹
 *  parameter1 hi2c:  I2CϥɥؤΥݥ
 *  parameter2 DevAddress: ǥХɥ쥹
 *  parameter3 MemAddress: ꥢɥ쥹
 *  parameter4 MemAddSize: ꥢɥ쥹
 *  parameter5 Timeout:    ॢȻ(ms)
 *  return                 ｪλE_OK
 */
static ER
i2c_masterreadaddress(I2C_Handle_t *hi2c, uint16_t DevAddress, uint16_t MemAddress, uint16_t MemAddSize, uint32_t Timeout)
{
	ER ercd = E_OK;

	/*
	 *  Υå
	 */
	sil_orw_mem((uint32_t *)(hi2c->base+TOFF_I2C_CR1), I2C_CR1_ACK);

	/*
	 *  Ԥ
	 */
	sil_orw_mem((uint32_t *)(hi2c->base+TOFF_I2C_CR1), I2C_CR1_START);
	if(i2c_sr1flagsetwait(hi2c, I2C_SR1_SB, Timeout) != E_OK){
		return E_TMOUT;
	}

	/*
	 *  ǥХɥ쥹Ԥ
	 */
	if(hi2c->Init.AddressingMode == I2C_ADDRESSINGMODE_7BIT){
		if(MemAddSize == 0)	/* ꥢɥ쥹ꤷʤ硢 */
			sil_wrw_mem((uint32_t *)(hi2c->base+TOFF_I2C_DR), I2C_7BIT_ADD_READ(DevAddress));
		else
			sil_wrw_mem((uint32_t *)(hi2c->base+TOFF_I2C_DR), I2C_7BIT_ADD_WRITE(DevAddress));
	}
	else{	/* ꥢɥ쥹ʤ */
		/*
		 *  10ӥåȥإåԤ
		 */
		sil_wrw_mem((uint32_t *)(hi2c->base+TOFF_I2C_DR), I2C_10BIT_HEADWRITE(DevAddress));
		if((ercd = i2c_addrflagwait(hi2c, I2C_SR1_ADD10, Timeout)) != E_OK){
			return ercd;
		}

		/*
		 *  10ӥåȥɥ쥹Ԥ
		 */
		sil_wrw_mem((uint32_t *)(hi2c->base+TOFF_I2C_DR), I2C_10BIT_ADDRESS(DevAddress));
		if((ercd = i2c_addrflagwait(hi2c, I2C_SR1_ADDR, Timeout)) != E_OK){
			return ercd;
		}

		/*
		 *  ɥ쥹ե饰ꥢ
		 */
		i2c_clear_addr(hi2c);

 		/*
		 *  ꥹԤ
		 */
		sil_orw_mem((uint32_t *)(hi2c->base+TOFF_I2C_CR1), I2C_CR1_START);
		if(i2c_sr1flagsetwait(hi2c, I2C_SR1_SB, Timeout) != E_OK){
			return E_TMOUT;
	    }

	   	/*
		 *  졼֥ɥ쥹⡼ɤ
		 */
		sil_wrw_mem((uint32_t *)(hi2c->base+TOFF_I2C_DR), I2C_10BIT_HEADREAD(DevAddress));
	}

	/*
	 *  ɥ쥹ե饰Ԥ
	 */
	if((ercd = i2c_addrflagwait(hi2c, I2C_SR1_ADDR, Timeout)) != E_OK){
		return ercd;
	}

	/*
	 *  ꥢɥ쥹ʤʤｪλ
	 */
	if(MemAddSize == 0)
		return E_OK;

	/*
	 *  ɥ쥹ե饰ꥢ
	 */
	i2c_clear_addr(hi2c);

	/*
	 *  TXEե饰ޤԤ
	 */
	if(i2c_sr1flagsetwait(hi2c, I2C_SR1_TXE, Timeout) != E_OK){
		return E_TMOUT;
	}

	/*
	 *  ꥢɥ쥹Ԥ
	 */
	if(MemAddSize == I2C_MEMADD_SIZE_8BIT){
		sil_wrw_mem((uint32_t *)(hi2c->base+TOFF_I2C_DR), I2C_MEM_ADD_LSB(MemAddress));
	}
	else{	/* 16ӥåȥ */
		sil_wrw_mem((uint32_t *)(hi2c->base+TOFF_I2C_DR), I2C_MEM_ADD_MSB(MemAddress));
		if(i2c_sr1flagsetwait(hi2c, I2C_SR1_TXE, Timeout) != E_OK){
			return E_TMOUT;
		}
		sil_wrw_mem((uint32_t *)(hi2c->base+TOFF_I2C_DR), I2C_MEM_ADD_LSB(MemAddress));
	}

	/*
	 *  TXEե饰ޤԤ
	 */
	if(i2c_sr1flagsetwait(hi2c, I2C_SR1_TXE, Timeout) != E_OK){
		return E_TMOUT;
	}

	/*
	 *  ꥹԤ
	 */
	sil_orw_mem((uint32_t *)(hi2c->base+TOFF_I2C_CR1), I2C_CR1_START);
	if(i2c_sr1flagsetwait(hi2c, I2C_SR1_SB, Timeout) != E_OK){
		return E_TMOUT;
	}

	/*
	 *  ǥХɥ쥹⡼ɤ(7ӥåȤΤ)
	 */
	sil_wrw_mem((uint32_t *)(hi2c->base+TOFF_I2C_DR), I2C_7BIT_ADD_READ(DevAddress));

	/*
	 *  ɥ쥹ե饰Ԥ
	 */
	return i2c_addrflagwait(hi2c, I2C_SR1_ADDR, Timeout);
}

/*
 *  ޥ ǥХɥ쥹ȥꥢɥ쥹
 *  parameter1 hi2c:  I2CϥɥؤΥݥ
 *  parameter2 DevAddress: ǥХɥ쥹
 *  parameter3 MemAddress: ꥢɥ쥹
 *  parameter4 MemAddSize: ꥢɥ쥹
 *  parameter5 Timeout:    ॢȻ(ms)
 *  return                 ｪλE_OK
 */
static ER
i2c_masterwriteaddress(I2C_Handle_t *hi2c, uint16_t DevAddress, uint16_t MemAddress, uint16_t MemAddSize, uint32_t Timeout)
{
	ER ercd = E_OK;

	/*
	 *  Ԥ
	 */
	sil_orw_mem((uint32_t *)(hi2c->base+TOFF_I2C_CR1), I2C_CR1_START);
	if(i2c_sr1flagsetwait(hi2c, I2C_SR1_SB, Timeout) != E_OK){
		return E_TMOUT;
	}

	/*
	 *  ǥХɥ쥹Ԥ
	 */
	if(hi2c->Init.AddressingMode == I2C_ADDRESSINGMODE_7BIT){
		sil_wrw_mem((uint32_t *)(hi2c->base+TOFF_I2C_DR), I2C_7BIT_ADD_WRITE(DevAddress));
	}
	else{
		sil_wrw_mem((uint32_t *)(hi2c->base+TOFF_I2C_DR), I2C_10BIT_HEADWRITE(DevAddress));
		if((ercd = i2c_addrflagwait(hi2c, I2C_SR1_ADD10, Timeout)) != E_OK){
			return ercd;
		}

		/*
		 *  10ӥåȥɥ쥹Ԥ
		 */
		sil_wrw_mem((uint32_t *)(hi2c->base+TOFF_I2C_DR), I2C_10BIT_ADDRESS(DevAddress));
	}

	/*
	 *  ɥ쥹ե饰Ԥ
	 */
	if((ercd = i2c_addrflagwait(hi2c, I2C_SR1_ADDR, Timeout)) != E_OK){
		return ercd;
	}

	/*
	 *  ɥ쥹ե饰ꥢ
	 */
	i2c_clear_addr(hi2c);

	/*
	 *  ꥢɥ쥹ʤʤｪλ
	 */
	if(MemAddSize == 0)
		return E_OK;

	/*
	 *  TXEե饰ޤԤ
	 */
	if(i2c_sr1flagsetwait(hi2c, I2C_SR1_TXE, Timeout) != E_OK){
		return E_TMOUT;
	}

	/*
	 *  ꥢɥ쥹Ԥ
	 */
	if(MemAddSize == I2C_MEMADD_SIZE_8BIT){
		sil_wrw_mem((uint32_t *)(hi2c->base+TOFF_I2C_DR), I2C_MEM_ADD_LSB(MemAddress));
	}
	else{	/* 16ӥåȥ */
		sil_wrw_mem((uint32_t *)(hi2c->base+TOFF_I2C_DR), I2C_MEM_ADD_MSB(MemAddress));
		if(i2c_sr1flagsetwait(hi2c, I2C_SR1_TXE, Timeout) != E_OK){
			return E_TMOUT;
		}
		sil_wrw_mem((uint32_t *)(hi2c->base+TOFF_I2C_DR), I2C_MEM_ADD_LSB(MemAddress));
	}
	return E_OK;
}

/*
 *  I2CǥХν
 *  parameter1 hi2c  I2CϥɥؤΥݥ
 *  return           ｪλE_OK
 */
I2C_Handle_t *
i2c_init(ID port, I2C_Init_t *ii2c)
{
	GPIO_Init_t GPIO_Init_Data;
	const I2C_PortControlBlock *ipcb;
	I2C_Handle_t *hi2c;
	uint32_t freqrange = 0;
	uint32_t no;

	if(port < I2C1_PORTID || port > NUM_I2CPORT)
		return NULL;
	if(ii2c == NULL)
		return NULL;

	no = INDEX_I2C(port);
	ipcb = &i2c_pcb[no];
	hi2c = &I2cHandle[no];
	if(hi2c->status != I2C_STATUS_RESET)
		return NULL;
	memcpy(&hi2c->Init, ii2c, sizeof(I2C_Init_t));
	hi2c->base  = ipcb->base;
	hi2c->i2cid = port;

	/*
	 *  SCL/SDAԥ/I2Cå
	 */
	sil_orw_mem((uint32_t *)ipcb->gioclockbase, ipcb->gioclockbit1);
	sil_orw_mem((uint32_t *)ipcb->gioclockbase, ipcb->gioclockbit2);
	sil_orw_mem((uint32_t *)ipcb->i2cclockbase, ipcb->i2cclockbit);

	GPIO_Init_Data.mode      = GPIO_MODE_AF;
	GPIO_Init_Data.pull      = GPIO_NOPULL;
	GPIO_Init_Data.otype     = GPIO_OTYPE_OD;
	GPIO_Init_Data.speed     = GPIO_SPEED_FAST;
	GPIO_Init_Data.alternate = ipcb->sclaf;
	gpio_setup(ipcb->giobase1, &GPIO_Init_Data, ipcb->sclpin);

	GPIO_Init_Data.alternate = ipcb->sdaaf;
	gpio_setup(ipcb->giobase2, &GPIO_Init_Data, ipcb->sdapin);


	hi2c->status = I2C_STATUS_BUSY;

	/*
	 *  I2C
	 */
	sil_andw_mem((uint32_t *)(hi2c->base+TOFF_I2C_CR1), I2C_CR1_PE);

	/*
	 * ܼȿ(PCLK1)󥸤
	 */
	freqrange = I2C_FREQRANGE(SysFrePCLK1);

	/*
	 *  I2Cȿ
	 */
	sil_wrw_mem((uint32_t *)(hi2c->base+TOFF_I2C_CR2), freqrange);

	/*
	 *  I2C饤
	 */
	sil_wrw_mem((uint32_t *)(hi2c->base+TOFF_I2C_TRISE), I2C_RISE_TIME(freqrange, hi2c->Init.ClockSpeed));

	/*
	 *  I2Cåԡ
	 */
	sil_wrw_mem((uint32_t *)(hi2c->base+TOFF_I2C_CCR), I2C_SPEED(SysFrePCLK1, hi2c->Init.ClockSpeed, hi2c->Init.DutyCycle));

	/*
	 *  I2C CR1
	 */
	sil_wrw_mem((uint32_t *)(hi2c->base+TOFF_I2C_CR1), (hi2c->Init.GeneralCallMode | hi2c->Init.NoStretchMode));

	/*
	 *  I2CǥХɥ쥹
	 */
	sil_wrw_mem((uint32_t *)(hi2c->base+TOFF_I2C_OAR1), (hi2c->Init.AddressingMode | hi2c->Init.OwnAddress1));

	/*
	 *  I2CǥХɥ쥹
	 */
	sil_wrw_mem((uint32_t *)(hi2c->base+TOFF_I2C_OAR2), (hi2c->Init.DualAddressMode | hi2c->Init.OwnAddress2));

	/*
	 *  I2Cư
	 */
	sil_orw_mem((uint32_t *)(hi2c->base+TOFF_I2C_CR1), I2C_CR1_PE);

	hi2c->ErrorCode = I2C_ERROR_NONE;
	hi2c->status = I2C_STATUS_READY;
	return hi2c;
}

/*
 *  I2CǥХ̵
 *  parameter1 hi2c  I2CϥɥؤΥݥ
 *  return           ｪλE_OK
 */
ER
i2c_deinit(I2C_Handle_t *hi2c)
{
	const I2C_PortControlBlock *ipcb;

	if(hi2c == NULL){
		return E_PAR;
	}

	hi2c->status = I2C_STATUS_BUSY;
	ipcb = &i2c_pcb[INDEX_I2C(hi2c->i2cid)];

	/*
	 *  I2C
	 */
	sil_andw_mem((uint32_t *)(hi2c->base+TOFF_I2C_CR1), I2C_CR1_PE);

	/*
	 *  I2Cåꥻå
	 */
	sil_orw_mem((uint32_t *)ipcb->i2cresetbase, ipcb->i2cresetbit);
	sil_andw_mem((uint32_t *)ipcb->i2cresetbase, ipcb->i2cresetbit);

	hi2c->ErrorCode = I2C_ERROR_NONE;
	hi2c->status = I2C_STATUS_RESET;
	return E_OK;
}

/*
 *  I2C졼ּ
 *  parameter1 hi2c:   I2CϥɥؤΥݥ
 *  parameter2 pData:  ɤ߽ФХåեؤΥݥ
 *  parameter3 Size:   ɤ߽Ф
 *  return             ｪλʤE_OK
 */
ER
i2c_slaveread(I2C_Handle_t *hi2c, uint8_t *pData, uint16_t Size)
{
	ER ercd = E_OK;

	if(pData == NULL || Size == 0)
		return E_PAR;

	/*
	 *  ɤ߽Фå
	 */
	if(hi2c->Init.semlock != 0)
		wai_sem(hi2c->Init.semlock);

	if(hi2c->status != I2C_STATUS_READY || i2c_busreadywait(hi2c, I2C_TIMEOUT_BUSY_FLAG) != E_OK){
		if(hi2c->Init.semlock != 0)
			sig_sem(hi2c->Init.semlock);
		return E_OBJ;
	}

	/*
	 *  Acknowledge/PEC Positionػ
	 */
	sil_andw_mem((uint32_t *)(hi2c->base+TOFF_I2C_CR1), I2C_CR1_POS);

    hi2c->status = I2C_STATUS_BUSY_RX;
    hi2c->ErrorCode = I2C_ERROR_NONE;

    hi2c->pBuffPtr = pData;
    hi2c->XferSize = Size;
    hi2c->XferCount = Size;

    /*
	 *  ɥ쥹Υå
	 */
	sil_orw_mem((uint32_t *)(hi2c->base+TOFF_I2C_CR1), I2C_CR1_ACK);

    /*
	 *  ɤ߽Фߵ
	 */
	sil_orw_mem((uint32_t *)(hi2c->base+TOFF_I2C_CR2), (I2C_CR2_ITEVTEN | I2C_CR2_ITBUFEN | I2C_CR2_ITERREN));
	ercd = i2c_transwait(hi2c, 500);

	/*
	 *  ɤ߽Фå
	 */
	if(hi2c->Init.semlock != 0)
		sig_sem(hi2c->Init.semlock);
	return ercd;
}

/*
 *  I2C졼
 *  parameter1 hi2c:   I2CϥɥؤΥݥ
 *  parameter2 pData:  ɤ߽ФХåեؤΥݥ
 *  parameter3 Size:   ɤ߽Ф
 *  return             ｪλʤE_OK
 */
ER
i2c_slavewrite(I2C_Handle_t *hi2c, uint8_t *pData, uint16_t Size)
{
	ER ercd = E_OK;

	if(pData == NULL || Size == 0)
		return E_PAR;

	/*
	 *  񤭹ߥå
	 */
	if(hi2c->Init.semlock != 0)
		wai_sem(hi2c->Init.semlock);

	if(hi2c->status != I2C_STATUS_READY || i2c_busreadywait(hi2c, I2C_TIMEOUT_BUSY_FLAG) != E_OK){
		if(hi2c->Init.semlock != 0)
			sig_sem(hi2c->Init.semlock);
		return E_OBJ;
	}

	/*
	 *  Acknowledge/PEC Positionػ
	 */
	sil_andw_mem((uint32_t *)(hi2c->base+TOFF_I2C_CR1), I2C_CR1_POS);

    hi2c->status = I2C_STATUS_BUSY_TX;
    hi2c->ErrorCode = I2C_ERROR_NONE;

    hi2c->pBuffPtr = pData;
    hi2c->XferSize = Size;
    hi2c->XferCount = Size;

    /*
	 *  ɥ쥹Υå
	 */
	sil_orw_mem((uint32_t *)(hi2c->base+TOFF_I2C_CR1), I2C_CR1_ACK);

    /*
	 *  񤭹߳ߵ
	 */
	sil_orw_mem((uint32_t *)(hi2c->base+TOFF_I2C_CR2), (I2C_CR2_ITEVTEN | I2C_CR2_ITBUFEN | I2C_CR2_ITERREN));
	ercd = i2c_transwait(hi2c, 500);

	/*
	 *  񤭹ߥå
	 */
	if(hi2c->Init.semlock != 0)
		sig_sem(hi2c->Init.semlock);
	return ercd;
}

/*
 *  I2Cǡ꡼
 *  parameter1 hi2c:       I2CϥɥؤΥݥ
 *  parameter2 DevAddress: 졼֥ɥ쥹
 *  parameter3 MemAddress: ꥢɥ쥹
 *  parameter4 MemAddSize: ꥢɥ쥹
 *  parameter5 pData:      ɤ߽ФХåեؤΥݥ
 *  parameter6 Size:       ɤ߽Ф
 *  return                 ｪλʤE_OK
 */
ER
i2c_memread(I2C_Handle_t *hi2c, uint16_t DevAddress, uint16_t MemAddress, uint16_t MemAddSize, uint8_t *pData, uint16_t Size)
{
	ER ercd = E_OK;

	if(pData == NULL || Size == 0)
		return E_PAR;
	if(hi2c->Init.AddressingMode != I2C_ADDRESSINGMODE_7BIT && MemAddSize != 0)
		return E_PAR;

	/*
	 *  ɤ߽Фå
	 */
	if(hi2c->Init.semlock != 0)
		wai_sem(hi2c->Init.semlock);

	if(hi2c->status != I2C_STATUS_READY || i2c_busreadywait(hi2c, I2C_TIMEOUT_BUSY_FLAG) != E_OK){
		if(hi2c->Init.semlock != 0)
			sig_sem(hi2c->Init.semlock);
		return E_OBJ;
	}

	/*
	 *  Acknowledge/PEC Positionػ
	 */
	sil_andw_mem((uint32_t *)(hi2c->base+TOFF_I2C_CR1), I2C_CR1_POS);

    hi2c->status = I2C_STATUS_BUSY_RX;
    hi2c->ErrorCode = I2C_ERROR_NONE;

    hi2c->pBuffPtr = pData;
    hi2c->XferSize = Size;
    hi2c->XferCount = Size;

	/* 졼֥ɥ쥹ɤ߽Фɥ쥹 */
	if((ercd = i2c_masterreadaddress(hi2c, DevAddress, MemAddress, MemAddSize, I2C_TIMEOUT_FLAG)) != E_OK){
		if(hi2c->Init.semlock != 0)
			sig_sem(hi2c->Init.semlock);
        return ercd;
	}

	if(hi2c->XferCount == 1){
		sil_andw_mem((uint32_t *)(hi2c->base+TOFF_I2C_CR1), I2C_CR1_ACK);
		i2c_clear_addr(hi2c);
		sil_orw_mem((uint32_t *)(hi2c->base+TOFF_I2C_CR1), I2C_CR1_STOP);
	}
	else if(hi2c->XferCount == 2){
		sil_andw_mem((uint32_t *)(hi2c->base+TOFF_I2C_CR1), I2C_CR1_ACK);
		sil_orw_mem((uint32_t *)(hi2c->base+TOFF_I2C_CR1), I2C_CR1_POS);
		i2c_clear_addr(hi2c);
	}
	else{
		sil_orw_mem((uint32_t *)(hi2c->base+TOFF_I2C_CR1), I2C_CR1_ACK);
		i2c_clear_addr(hi2c);
	}

    /*
	 *  ɤ߽Фߵ
	 */
	sil_orw_mem((uint32_t *)(hi2c->base+TOFF_I2C_CR2), (I2C_CR2_ITEVTEN | I2C_CR2_ITBUFEN | I2C_CR2_ITERREN));
	ercd = i2c_transwait(hi2c, 100);

	/*
	 *  ɤ߽Фå
	 */
	if(hi2c->Init.semlock != 0)
		sig_sem(hi2c->Init.semlock);
	return ercd;
}

/*
 *  I2Cǡ饤
 *  parameter1 hi2c:       I2CϥɥؤΥݥ
 *  parameter2 DevAddress: 졼֥ɥ쥹
 *  parameter3 MemAddress: ꥢɥ쥹
 *  parameter4 MemAddSize: ꥢɥ쥹
 *  parameter5 pData:      ߥХåեؤΥݥ
 *  parameter6 Size:       ߥ
 *  return                 ｪλʤE_OK
 */
ER
i2c_memwrite(I2C_Handle_t *hi2c, uint16_t DevAddress, uint16_t MemAddress, uint16_t MemAddSize, uint8_t *pData, uint16_t Size)
{
	ER ercd = E_OK;

	if(pData == NULL || Size == 0)
		return  E_PAR;
	if(hi2c->Init.AddressingMode != I2C_ADDRESSINGMODE_7BIT && MemAddSize != 0)
		return  E_PAR;

	/*
	 *  񤭹ߥå
	 */
	if(hi2c->Init.semlock != 0)
		wai_sem(hi2c->Init.semlock);

	if(hi2c->status != I2C_STATUS_READY || i2c_busreadywait(hi2c, I2C_TIMEOUT_BUSY_FLAG) != E_OK){
		if(hi2c->Init.semlock != 0)
			sig_sem(hi2c->Init.semlock);
		return E_OBJ;
	}

	/*
	 *  Acknowledge/PEC Positionػ
	 */
	sil_andw_mem((uint32_t *)(hi2c->base+TOFF_I2C_CR1), I2C_CR1_POS);

	hi2c->status = I2C_STATUS_BUSY_TX;
	hi2c->ErrorCode = I2C_ERROR_NONE;

	hi2c->pBuffPtr = pData;
	hi2c->XferSize = Size;
	hi2c->XferCount = Size;

	/* 졼֥ɥ쥹Ƚߥɥ쥹 */
	if((ercd = i2c_masterwriteaddress(hi2c, DevAddress, MemAddress, MemAddSize, I2C_TIMEOUT_FLAG)) != E_OK){
		if(hi2c->Init.semlock != 0)
			sig_sem(hi2c->Init.semlock);
		return ercd;
	}

    /*
	 *  񤭹߳ߵ
	 */
	sil_orw_mem((uint32_t *)(hi2c->base+TOFF_I2C_CR2), (I2C_CR2_ITEVTEN | I2C_CR2_ITBUFEN | I2C_CR2_ITERREN));
	ercd = i2c_transwait(hi2c, 500);

	/*
	 *  񤭹ߥå
	 */
	if(hi2c->Init.semlock != 0)
		sig_sem(hi2c->Init.semlock);
	return ercd;
}

/*
 *  I2Cǥå
 *  parameter1 hi2c:       I2CϥɥؤΥݥ
 *  parameter2 DevAddress: 졼֥ɥ쥹
 *  parameter3 Trials:     ȥ饤
 *  parameter4 Timeout:    ॢ
 *  return                 ǥʤE_OK
 */
ER
i2c_ready(I2C_Handle_t *hi2c, uint16_t DevAddress, uint32_t Trials, uint32_t Timeout)
{
	volatile uint32_t I2C_Trials = 0;
	uint32_t tmp1 = 0, tmp2 = 0;
	uint32_t tick = 0;

	if(hi2c->status == I2C_STATUS_READY){
		if((sil_rew_mem((uint32_t *)(hi2c->base+TOFF_I2C_SR2)) & I2C_SR2_BUSY) != 0)
			return E_OBJ;

		/*
		 *  å
		 */
		if(hi2c->Init.semlock != 0)
			wai_sem(hi2c->Init.semlock);

		sil_andw_mem((uint32_t *)(hi2c->base+TOFF_I2C_CR1), I2C_CR1_POS);
		hi2c->status    = I2C_STATUS_BUSY;
		hi2c->ErrorCode = I2C_ERROR_NONE;

		do{
			/* START */
			sil_orw_mem((uint32_t *)(hi2c->base+TOFF_I2C_CR1), I2C_CR1_START);

			/* START BITԤ */
			if(i2c_sr1flagsetwait(hi2c, I2C_SR1_SB, Timeout) != E_OK){
				/*
				 *  å
				 */
				if(hi2c->Init.semlock != 0)
					sig_sem(hi2c->Init.semlock);
				return E_TMOUT;
			}

			/* 졼֥ɥ쥹 */
			sil_wrw_mem((uint32_t *)(hi2c->base+TOFF_I2C_DR), I2C_7BIT_ADD_WRITE(DevAddress));

			/* ɥ쥹ΥåԤ */
			tmp1 = sil_rew_mem((uint32_t *)(hi2c->base+TOFF_I2C_SR1)) & (I2C_SR1_ADDR | I2C_SR1_AF);
			tmp2 = hi2c->status;
			while(tmp1 == 0 && tmp2 != I2C_STATUS_TIMEOUT){
				if(tick > Timeout){
					hi2c->status = I2C_STATUS_TIMEOUT;
				}
				tmp1 = sil_rew_mem((uint32_t *)(hi2c->base+TOFF_I2C_SR1)) & (I2C_SR1_ADDR | I2C_SR1_AF);
				tmp2 = hi2c->status;
				dly_tsk(1);
				tick++;
			}
			hi2c->status = I2C_STATUS_READY;

			/* ɥ쥹ե饰å */
			if((sil_rew_mem((uint32_t *)(hi2c->base+TOFF_I2C_SR1)) & I2C_SR1_ADDR) != 0){
				/* ȥå */
				sil_orw_mem((uint32_t *)(hi2c->base+TOFF_I2C_CR1), I2C_CR1_STOP);

				/* ɥ쥹ե饰ꥢ */
				tmp1 = sil_rew_mem((uint32_t *)(hi2c->base+TOFF_I2C_SR1));
				tmp1 = sil_rew_mem((uint32_t *)(hi2c->base+TOFF_I2C_SR2));

				/* ǥԤ */
				if(i2c_busreadywait(hi2c, I2C_TIMEOUT_BUSY_FLAG) != E_OK){
					/*
					 *  å
					 */
					if(hi2c->Init.semlock != 0)
						sig_sem(hi2c->Init.semlock);
					return E_TMOUT;
				}

				hi2c->status = I2C_STATUS_READY;
				/*
				 *  å
				 */
				if(hi2c->Init.semlock != 0)
					sig_sem(hi2c->Init.semlock);
				return E_OK;
			}
			else{	/* Υå */
				/* ȥå */
				sil_orw_mem((uint32_t *)(hi2c->base+TOFF_I2C_CR1), I2C_CR1_STOP);

				/* ɥ쥹ե饰ꥻå */
				sil_andw_mem((uint32_t *)(hi2c->base+TOFF_I2C_SR1), I2C_SR1_ADDR);

				/* ǥԤ */
				if(i2c_busreadywait(hi2c, I2C_TIMEOUT_BUSY_FLAG) != E_OK){
					/*
					 *  å
					 */
					if(hi2c->Init.semlock != 0)
						sig_sem(hi2c->Init.semlock);
					return E_TMOUT;
				}
			}
		}while(I2C_Trials++ < Trials);

		hi2c->status = I2C_STATUS_READY;
		/*
		 *  å
		 */
		if(hi2c->Init.semlock != 0)
			sig_sem(hi2c->Init.semlock);
	    return E_TMOUT;
	}
	else
		return E_OBJ;
}

/*
 *  I2CХ
 *  parameter1 hi2c:       I2CϥɥؤΥݥ
 */
static void i2c_transmit_byte(I2C_Handle_t *hi2c)
{
	if(hi2c->XferCount != 0){
		sil_wrw_mem((uint32_t *)(hi2c->base+TOFF_I2C_DR), (uint32_t)(*hi2c->pBuffPtr++));
		hi2c->XferCount--;
	}
}

/*
 *  I2CХȼ
 *  parameter1 hi2c:       I2CϥɥؤΥݥ
 */
static void i2c_receive_byte(I2C_Handle_t *hi2c)
{
	if(hi2c->XferCount != 0){
		(*hi2c->pBuffPtr++) = (uint8_t)sil_rew_mem((uint32_t *)(hi2c->base+TOFF_I2C_DR));
		hi2c->XferCount--;
	}
}

/*
 *  I2Cޥ(BTF)
 */
static void
i2c_mastertrans_BTF(I2C_Handle_t *hi2c)
{
	if(hi2c->XferCount != 0){	/* ǡ */
		sil_wrw_mem((uint32_t *)(hi2c->base+TOFF_I2C_DR), (uint32_t)(*hi2c->pBuffPtr++));
		hi2c->XferCount--;
	}
	else{	/* λ */
		sil_andw_mem((uint32_t *)(hi2c->base+TOFF_I2C_CR2), (I2C_CR2_ITEVTEN | I2C_CR2_ITBUFEN | I2C_CR2_ITERREN));
		sil_orw_mem((uint32_t *)(hi2c->base+TOFF_I2C_CR1), I2C_CR1_STOP);

		hi2c->status = I2C_STATUS_READY;
		if(hi2c->writecallback != NULL)
			hi2c->writecallback(hi2c);
	}
}


/*
 *  I2Cޥ(TXE)
 */
static void
i2C_MasterTransmit_TXE(I2C_Handle_t *hi2c)
{
	/*
	 *  Х
	 */
	sil_wrw_mem((uint32_t *)(hi2c->base+TOFF_I2C_DR), (uint32_t)(*hi2c->pBuffPtr++));
	hi2c->XferCount--;

	if(hi2c->XferCount == 0){	/*  */
		sil_andw_mem((uint32_t *)(hi2c->base+TOFF_I2C_CR2), I2C_CR2_ITBUFEN);
	}
}

/*
 *  I2Cޥ(RXNE)
 */
static void
i2c_masterreceiv_RXNE(I2C_Handle_t *hi2c)
{
	uint32_t tmp = 0;

	tmp = hi2c->XferCount;
	if(tmp > 3){	/* Ĥ4ХȰʾ */
		(*hi2c->pBuffPtr++) = (uint8_t)sil_rew_mem((uint32_t *)(hi2c->base+TOFF_I2C_DR));
		hi2c->XferCount--;
	}
	else if(tmp == 2 || tmp == 3){	/* Ĥ2ޤ3Х */
		sil_andw_mem((uint32_t *)(hi2c->base+TOFF_I2C_CR2), I2C_CR2_ITBUFEN);
	}
	else{	/* λ */
		sil_andw_mem((uint32_t *)(hi2c->base+TOFF_I2C_CR2), (I2C_CR2_ITEVTEN | I2C_CR2_ITBUFEN | I2C_CR2_ITERREN));

		(*hi2c->pBuffPtr++) = (uint8_t)sil_rew_mem((uint32_t *)(hi2c->base+TOFF_I2C_DR));
		hi2c->XferCount--;

		hi2c->status = I2C_STATUS_READY;
		if(hi2c->readcallback != NULL)
			hi2c->readcallback(hi2c);
	}
}

/*
 *  I2Cޥ(BTF)
 */
static void
i2c_masterreceiv_BTF(I2C_Handle_t *hi2c)
{
	if(hi2c->XferCount == 3){		/* Ĥꥫ3Х */
		sil_andw_mem((uint32_t *)(hi2c->base+TOFF_I2C_CR1), I2C_CR1_ACK);
		(*hi2c->pBuffPtr++) = (uint8_t)sil_rew_mem((uint32_t *)(hi2c->base+TOFF_I2C_DR));
		hi2c->XferCount--;
	}
	else if(hi2c->XferCount == 2){	/* Ĥꥫ2Х */
		sil_orw_mem((uint32_t *)(hi2c->base+TOFF_I2C_CR1), I2C_CR1_STOP);
		(*hi2c->pBuffPtr++) = (uint8_t)sil_rew_mem((uint32_t *)(hi2c->base+TOFF_I2C_DR));
		hi2c->XferCount--;
		(*hi2c->pBuffPtr++) = (uint8_t)sil_rew_mem((uint32_t *)(hi2c->base+TOFF_I2C_DR));
		hi2c->XferCount--;

		/*
		 *  
		 */
		sil_andw_mem((uint32_t *)(hi2c->base+TOFF_I2C_CR2), (I2C_CR2_ITEVTEN | I2C_CR2_ITERREN));

		hi2c->status = I2C_STATUS_READY;
		if(hi2c->readcallback != NULL)
			hi2c->readcallback(hi2c);
	}
	else{
		(*hi2c->pBuffPtr++) = (uint8_t)sil_rew_mem((uint32_t *)(hi2c->base+TOFF_I2C_DR));
		hi2c->XferCount--;
	}
}

/*
 *  I2C EVߥϥɥ
 */
void
i2c_ev_handler(I2C_Handle_t *hi2c)
{
	uint32_t sr1 = sil_rew_mem((uint32_t *)(hi2c->base+TOFF_I2C_SR1));
	uint32_t sr2 = sil_rew_mem((uint32_t *)(hi2c->base+TOFF_I2C_SR2));
	uint32_t cr2 = sil_rew_mem((uint32_t *)(hi2c->base+TOFF_I2C_CR2));
	volatile uint32_t tmp;

	/*
	 *  ޥ⡼
	 */
	if((sr2 & I2C_SR2_MSL) != 0){
		if((sr2 & I2C_SR2_TRA) != 0){	/* ⡼ */
			if(((sr1 & I2C_SR1_TXE) != 0) && ((cr2 & I2C_CR2_ITBUFEN) != 0) && ((sr1 & I2C_SR1_BTF) == 0)){
				i2C_MasterTransmit_TXE(hi2c);
			}
			else if(((sr1 & I2C_SR1_BTF) != 0) && ((cr2 & I2C_CR2_ITEVTEN) != 0)){
				i2c_mastertrans_BTF(hi2c);
			}
		}
		else{							/* ⡼ */
			if(((sr1 & I2C_SR1_RXNE) != 0) && ((cr2 & I2C_CR2_ITBUFEN) != 0) && ((sr1 & I2C_SR1_BTF) == 0)){
				i2c_masterreceiv_RXNE(hi2c);
			}
 			else if(((sr1 & I2C_SR1_BTF) != 0) && ((cr2 & I2C_CR2_ITEVTEN) != 0)){
				i2c_masterreceiv_BTF(hi2c);
			}
		}
		if(hi2c->status == I2C_STATUS_READY && hi2c->Init.semid != 0)
			isig_sem(hi2c->Init.semid);
	}
	/*
	 *  졼֥⡼
	 */
	else{
		if(((sr1 & I2C_SR1_ADDR) != 0) && ((cr2 & I2C_CR2_ITEVTEN) != 0)){	/* ADDRե饰 */
			i2c_clear_addr(hi2c);
		}
		else if(((sr1 & I2C_SR1_STOPF) != 0) && ((cr2 & I2C_CR2_ITEVTEN) != 0)){	/* STOPե饰 */
			sil_andw_mem((uint32_t *)(hi2c->base+TOFF_I2C_CR2), (I2C_CR2_ITEVTEN | I2C_CR2_ITBUFEN | I2C_CR2_ITERREN));
			/*
			 * STOPե饰ꥢ
			 */
			tmp = sil_rew_mem((uint32_t *)(hi2c->base+TOFF_I2C_SR1));
			sil_orw_mem((uint32_t *)(hi2c->base+TOFF_I2C_CR1), I2C_CR1_PE);
			(void)(tmp);

		    /*
			 *  ɥ쥹Υåػ
			 */
			sil_andw_mem((uint32_t *)(hi2c->base+TOFF_I2C_CR1), I2C_CR1_ACK);

			hi2c->status = I2C_STATUS_READY;
			if(hi2c->readcallback != NULL)
				hi2c->readcallback(hi2c);
		}
		else if((sr2 & I2C_SR2_TRA) != 0){	/* 졼 */
			if(((sr1 & I2C_SR1_TXE) != 0) && ((cr2 & I2C_CR2_ITBUFEN) != 0) && ((sr1 & I2C_SR1_BTF) == 0)){
				i2c_transmit_byte(hi2c);
			}
			else if(((sr1 & I2C_SR1_BTF) != 0) && ((cr2 & I2C_CR2_ITEVTEN) != 0)){
				i2c_transmit_byte(hi2c);
			}
		}
		else{								/* 졼ּ */
			if(((sr1 & I2C_SR1_RXNE) != 0) && ((cr2 & I2C_CR2_ITBUFEN) != 0) && ((sr1 & I2C_SR1_BTF) == 0)){
				i2c_receive_byte(hi2c);
			}
			else if(((sr1 & I2C_SR1_BTF) != 0) && ((cr2 & I2C_CR2_ITEVTEN) != 0)){
				i2c_receive_byte(hi2c);
			}
		}
	}
}

/*
 *  I2C ERߥϥɥ
 */
void
i2c_er_handler(I2C_Handle_t *hi2c)
{
	uint32_t sr1 = sil_rew_mem((uint32_t *)(hi2c->base+TOFF_I2C_SR1));
	uint32_t cr2 = sil_rew_mem((uint32_t *)(hi2c->base+TOFF_I2C_CR2));

	/*
	 *  I2C Х顼
	 */
	if(((sr1 & I2C_SR1_BERR) != 0) && ((cr2 & I2C_CR2_ITERREN) != 0)){
		hi2c->ErrorCode |= I2C_ERROR_BERR;
		sil_wrw_mem((uint32_t *)(hi2c->base+TOFF_I2C_SR1), ~I2C_SR1_BERR);
	}
	sr1 = sil_rew_mem((uint32_t *)(hi2c->base+TOFF_I2C_SR1));

	/*
	 *  I2C ӥơ󡦥顼
	 */
	if(((sr1 & I2C_SR1_ARLO) != 0) && ((cr2 & I2C_CR2_ITERREN) != 0)){
		hi2c->ErrorCode |= I2C_ERROR_ARLO;
		sil_wrw_mem((uint32_t *)(hi2c->base+TOFF_I2C_SR1), ~I2C_SR1_ARLO);
	}
	sr1 = sil_rew_mem((uint32_t *)(hi2c->base+TOFF_I2C_SR1));

	/*
	 *  I2C Υåե롦顼
	 */
	if(((sr1 & I2C_SR1_AF) != 0) && ((cr2 & I2C_CR2_ITERREN) != 0)){
		uint32_t sr2 = sil_rew_mem((uint32_t *)(hi2c->base+TOFF_I2C_SR2));
		if(((sr2 & I2C_SR2_MSL) == 0) && (hi2c->XferCount == 0) && (hi2c->status == I2C_STATUS_BUSY_TX)){
			/* Disable EVT, BUF and ERR interrupt */
			sil_andw_mem((uint32_t *)(hi2c->base+TOFF_I2C_CR2), (I2C_CR2_ITEVTEN | I2C_CR2_ITBUFEN | I2C_CR2_ITERREN));
			/* Clear AF flag */
			sil_wrw_mem((uint32_t *)(hi2c->base+TOFF_I2C_SR1), ~I2C_SR1_AF);
			/* Disable Acknowledge */
			sil_andw_mem((uint32_t *)(hi2c->base+TOFF_I2C_CR1), I2C_CR1_ACK);
			hi2c->status = I2C_STATUS_READY;
			if(hi2c->writecallback != NULL)
				hi2c->writecallback(hi2c);
		}
		else{
			hi2c->ErrorCode |= I2C_ERROR_AF;
			sil_wrw_mem((uint32_t *)(hi2c->base+TOFF_I2C_SR1), ~I2C_SR1_AF);
		}
	}
	sr1 = sil_rew_mem((uint32_t *)(hi2c->base+TOFF_I2C_SR1));

	/*
	 *  I2C С/󡦥顼
	 */
	if(((sr1 & I2C_SR1_OVR) != 0/*tmp1 == SET*/) && ((cr2 & I2C_CR2_ITERREN) != 0)){
		hi2c->ErrorCode |= I2C_ERROR_OVR;
		sil_wrw_mem((uint32_t *)(hi2c->base+TOFF_I2C_SR1), ~I2C_SR1_OVR);
	}

	if(hi2c->ErrorCode != I2C_ERROR_NONE){
		hi2c->status = I2C_STATUS_READY;

		/*
		 *  Acknowledge/PEC Positionػ
		 */
		sil_andw_mem((uint32_t *)(hi2c->base+TOFF_I2C_CR1), I2C_CR1_POS);

		if(hi2c->errorcallback != NULL)
			hi2c->errorcallback(hi2c);
	}
}


/*
 *  I2C EVߥӥ롼
 */
void
i2c_ev_isr(intptr_t exinf)
{
	i2c_ev_handler(&I2cHandle[INDEX_I2C((uint32_t)exinf)]);
}

/*
 *  I2C ERߥӥ롼
 */
void
i2c_er_isr(intptr_t exinf)
{
	i2c_er_handler(&I2cHandle[INDEX_I2C((uint32_t)exinf)]);
}

